/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <command.h>

static command_t *retrieve_command(char *cmd, ptable_t *fixed_cmds);

static int authenticator_cmd(int argc, char **argv);
static int buffer_size_cmd(int argc, char **argv);
static int connect_cmd(int argc, char **argv);
static int device_cmd(int argc, char **argv);
static int dev_ip_cmd(int argc, char **argv);
static int dns_limit_cmd(int argc, char **argv);
static int dns_priority_cmd(int argc, char **argv);
static int group_cmd(int argc, char **argv);
static int help_cmd(int argc, char **argv);
static int limit_cmd(int argc, char **argv);
static int load_cmd(int argc, char **argv);
static int log_file_cmd(int argc, char **argv);
static int log_level_cmd(int argc, char **argv);
static int persist_cmd(int argc, char **argv);
static int plugin_dir_cmd(int argc, char **argv);
static int server_cmd(int argc, char **argv);
static int timeout_cmd(int argc, char **argv);
static int user_cmd(int argc, char **argv);

static const command_t internal_cmds[] = {
	{ "authenticator", CMD_EXTERNAL, authenticator_cmd, "authenticator <authenticator type> - Set current authenticator used in server/connect. Can be changed each time before those command." },
	{ "buffer-size", CMD_INTERNAL, buffer_size_cmd, "buffer-size <size> - Set internal buffers size, use a kilo bytes factor (e.g.: 16 will be 16 * 1024 = 16kb etc...). Default is 16." },
	{ "connect", CMD_EXTERNAL,connect_cmd, "connect <connector:cypher:host<:port>> - Try to connect to ip and port (default one if not given) with connector." },
	{ "device", CMD_INTERNAL, device_cmd, "device <name> - set TUN device, and optionnaly its name. (by default it will take the one given by the driver, something like tun<nb>)." },
	{ "dev-ip", CMD_INTERNAL, dev_ip_cmd, "dev-ip <ip|uip>/<prefix length> - set TUN device IP (link local or global) address." },
	{ "dns-limit", CMD_INTERNAL, dns_limit_cmd, "dns-limit <limit> - set concurrent DNS request limit, should be between 1-16. (default is 16)" },
	{ "dns-priority", CMD_EXTERNAL, dns_priority_cmd, "dns-priority <ipv4/ipv6> - set DNS request network family priority. (default is ipv4)" },
	{ "group", CMD_INTERNAL, group_cmd, "group <group name> - Set the group's rights it will run with." },
	{ "help", CMD_EXTERNAL, help_cmd, "help <command> - displays command help, or command list if no argument is given." },
	{ "limit", CMD_EXTERNAL, limit_cmd, "limit <num> - limit the number of possible mnode connections." },
	{ "load", CMD_EXTERNAL, load_cmd, "load <plugin name> - load one plugin identified by its name." },
	{ "log-file", CMD_INTERNAL, log_file_cmd, "log-file <log file path> - set the log file." },
	{ "log-level", CMD_EXTERNAL, log_level_cmd, "log-level <log level> - set logger verbosity (0-3)." },
	{ "persist", CMD_INTERNAL, persist_cmd, "persist <0|1>- set TUN device persistent (or not if already persistent)." },
	{ "plugin-dir", CMD_EXTERNAL, plugin_dir_cmd, "plugin_dir <dir path> - set where mnode will look for loading plugins." },
	{ "server", CMD_EXTERNAL, server_cmd, "server <connector:authcypher:ip<:port>> - set connection listener for connector at ip and port (default one if not given)." },
	{ "timeout", CMD_EXTERNAL, timeout_cmd, "timeout <seconds> - Set the timeout in all connections i/o, should be between 10-600. (default is 30 seconds)" },
	{ "user", CMD_INTERNAL, user_cmd, "user <username> - Set the user's rights it will run with." },
	{ "", CMD_NONE, NULL, "" }
};

ptable_t *command_set = NULL;

extern char plugin_dir[512];
extern mdevice_t *mdevice_g;

char **parse_command(char *cmd_line)
{
	char *c_pos = NULL;
	char *s_pos = NULL;
	char **cmd = NULL;
	char **cmd_r = NULL;
	int line_len = 0;
	int size_cpy = 0;
	int len_minus;
	int n_param = 1;

	if((c_pos = strchr(cmd_line, '\n')) != NULL)
		*c_pos = '\0';

	if((c_pos = strchr(cmd_line, '\r')) != NULL)
		*c_pos = '\0';

	if ((c_pos = strchr(cmd_line, '#')) != NULL)
		*c_pos = '\0';

	for (c_pos = cmd_line; *c_pos == ' '; c_pos++);

	line_len = strlen(c_pos);
	for (; line_len > 0 && n_param <= COMMAND_MAX_PARAMS; 
		 n_param++, c_pos = s_pos+1, line_len -= len_minus)
	{
		len_minus = 0;
		for (; *c_pos == ' '; c_pos++, len_minus++);

		s_pos = c_pos;
		if (*c_pos == '"')
		{
			c_pos++;
			len_minus++;
			s_pos = strchr(c_pos, '"');
			if (s_pos == NULL)
			{
				free_cmd_line(cmd);
				return NULL;
			}
		}
		else
			s_pos = strchr(s_pos, ' ');

		size_cpy = (s_pos == NULL? line_len : (int) (s_pos - c_pos));
		len_minus += size_cpy+1;

		cmd_r = (char **) realloc(cmd, sizeof(char *) * (int) (n_param + 1));
		if (cmd_r == NULL)
		{
			ERR("parse_command()");
			free_cmd_line(cmd);
			return NULL;
		}

		cmd = cmd_r;
		cmd[n_param] = NULL;

		cmd[n_param - 1] = (char *) calloc(size_cpy + 1, sizeof(char));
		if (cmd[n_param - 1] == NULL)
		{
			ERR("parse_command()");
			free_cmd_line(cmd);
			return NULL;
		}

		strncpy(cmd[n_param - 1], c_pos, size_cpy);
	}

	return cmd;
}

void free_cmd_line(char **cmd)
{
	int cmd_idx;

	if (cmd == NULL)
		return;

	for(cmd_idx = 0; cmd[cmd_idx] != NULL; cmd_idx++)
		free(cmd[cmd_idx]);

	free(cmd);
}

int init_commands(void)
{
	command_set = init_command_set(internal_cmds);
	if (command_set == NULL)
		return -1;

	return 0;
}

ptable_t *init_command_set(const command_t *cmd_set)
{
	ptable_t *p_cmd = NULL;
	int idx;

	p_cmd = pt_create(PTABLE_SEARCH, 0, command_cmp);
	if (p_cmd == NULL)
	{
		log_msg(ERROR_LVL, "p_cmd not initialized\n");
		return NULL;
	}

	log_msg(INFO_LVL, "Inserting commands\n");
	for (idx = 0; cmd_set[idx].command != NULL; idx++)
	{
		if (pt_insert(p_cmd, (void *) &(cmd_set[idx])) < 0)
		{
			pt_destroy(p_cmd);
			return NULL;
		}
	}

	return p_cmd;
}

int interpret_command(char **cmd, cmd_level_t level, ptable_t *fixed_cmds)
{
	command_t *cmd_run = NULL;
	int cnt;

	if (cmd == NULL)
		return CMD_UNKNOWN;

	cmd_run = retrieve_command(cmd[0], fixed_cmds);

	if (cmd_run  == NULL)
		return CMD_UNKNOWN;

	if (cmd_run->level > level)
		return CMD_UNKNOWN;

	for(cnt = 0; cmd[cnt] != NULL; cnt++);

	return cmd_run->command(cnt, cmd);
}

int command_cmp(const void *c1, const void *c2)
{
	command_t *cmd_1 = (command_t *) *((command_t **) c1);
	command_t *cmd_2 = (command_t *) *((command_t **) c2);

	return strncmp(cmd_1->name, cmd_2->name, COMMAND_NAME_MAX_LENGTH);
}

int add_command(command_t *cmd)
{
	if (cmd == NULL)
		return -1;

	if (retrieve_command(cmd->name, command_set) != NULL)
		return -1;

	if (pt_insert(command_set, (void *) cmd) < 0)
		return -1;

	return 0;
}

int del_command(command_t *cmd)
{
	int idx;

	if (cmd == NULL)
		return -1;

	if (retrieve_command(cmd->name, command_set) == NULL)
		return 0;

	for (idx = 0; (idx < command_set->count) && (command_set->ptable[idx] != cmd); idx++);

	if (idx >= command_set->count)
		return -1;

	if (pt_remove(command_set, idx) == NULL)
		return -1;

	return 0;
}

static command_t *retrieve_command(char *cmd, ptable_t *fixed_cmds)
{
	command_t cmd_test;
	ptable_t *cmd_set = command_set;

	if (cmd == NULL)
		return NULL;

	memset(&cmd_test, 0, sizeof(command_t));
	strncpy(cmd_test.name, cmd, COMMAND_NAME_MAX_LENGTH);

	if (fixed_cmds != NULL)
		cmd_set = fixed_cmds;

	return pt_search(cmd_set, &cmd_test);
}

/**********\ 
* COMMANDS * 
\**********/

static int authenticator_cmd(int argc, char **argv)
{
	if (argc != 2)
	{
		log_msg(ERROR_LVL, "authenticator: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	mdevice_g->authenticator_id = get_authenticator_id(argv[1]);
	if (mdevice_g->authenticator_id < 0)
		log_msg(WARNING_LVL, "authenticator: no such type %s. Setting no authenticator.\n", argv[1]);

	return CMD_SUCCESS;
}

static int buffer_size_cmd(int argc, char **argv)
{
	short val;
	uint8_t *buff = NULL;

	if (argc != 2)
	{
		log_msg(ERROR_LVL, "buffer-size: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	errno = 0;
	val = (short) strtol(argv[1], (char **)NULL, 10);
	if ((errno == ERANGE) || (val <= 0))
	{
		log_msg(ERROR_LVL, "buffer-size: bad number\n");
		return CMD_ERROR;
	}

	if ((val * 1024) != mdevice_g->buffer_size)
	{
		buff = (uint8_t *) calloc(val * 1024, sizeof(uint8_t));
		if (buff == NULL)
		{
			ERR("buffer_size_cmd()");
			log_msg(ERROR_LVL, "buffer-size: could not change buffer size\n");
			return CMD_ERROR;
		}
		
	    free(mdevice_g->w_buffer);
		mdevice_g->w_buffer = buff;
		mdevice_g->buffer_size = val * 1024;
	}

	return CMD_SUCCESS;
}

static int connect_cmd(int argc, char **argv)
{
	mnode_t *new_mnode = NULL;

	if (argc != 2)
	{
		log_msg(ERROR_LVL, "connect: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	if ((new_mnode = init_mnode()) == NULL)
	{
		log_msg(ERROR_LVL, "connect: could not get memory for a new mnode\n");
		return CMD_ERROR;
	}

	if (parse_connector_line(argv[1], &new_mnode->connector_id, 
							 &new_mnode->cipher_id, 
							 new_mnode->host, &new_mnode->port) < 0)
	{
		log_msg(ERROR_LVL, "connect line badly written: %s\n", argv[1]);
		free_mnode(new_mnode);
		return CMD_ERROR;
	}


	DEBUG_MSG("connect_cmd(): mnode connector_id %d and cipher_id %d\n", 
			  new_mnode->connector_id, new_mnode->cipher_id);

	if (hostname_lookup(mdevice_g, new_mnode->host, mnode_dns_resolution, new_mnode) < 0)
	{
		log_msg(ERROR_LVL, "Could not proceed in hostname resolution\n");
		free_mnode(new_mnode);
		return CMD_ERROR;
	}

	return CMD_SUCCESS;
}

static int device_cmd(int argc, char **argv)
{
	if (argc > 2)
	{
		log_msg(ERROR_LVL, "device: takes at most 1 argument.\n");
		return CMD_ERROR;
	}

	if (argc == 2)
	{
		if (strlen(argv[1]) > IF_NAMESIZE)
		{
			log_msg(ERROR_LVL, "device: name \"%s\" is too long for a device name. Default name will be used\n");
			return CMD_ERROR;
		}

		strncpy(mdevice_g->dev_name, argv[1], IF_NAMESIZE);
	}

	if (open_tun_device(mdevice_g) < 0)
	{
		log_msg(FATAL_LVL, "device: could not open TUN device - exiting\n");
		exit(EXIT_FAILURE);
	}

	if (set_device_name(mdevice_g) < 0)
		log_msg(WARNING_LVL, "device: could not set/get device name, trying to continue...\n");

	if (set_device_up(mdevice_g, true) < 0)
		log_msg(WARNING_LVL, "device: could not set device up, trying to continue...\n");

	return CMD_SUCCESS;
}

static int dev_ip_cmd(int argc, char **argv)
{
	char address[INET6_ADDRSTRLEN];
	int prefix;
	struct in6_addr ip6_addr;

	if (argc != 2)
	{
		log_msg(ERROR_LVL, "dev-ip: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	if (parse_address_line(argv[1], address, &prefix) < 0)
	{
		log_msg(ERROR_LVL, "dev-ip line badly written: %s - check address/prefix\n", argv[1]);
		return CMD_ERROR;
	}

	if (inet_pton(AF_INET6, argv[1], &ip6_addr) <= 0)
	{
		log_msg(ERROR_LVL, "vpn-ip: could not set the vpn ip %s\n", argv[1]);
		return CMD_ERROR;
	}

	if (set_device_address(mdevice_g, ip6_addr, prefix) < 0)
	{
		log_msg(ERROR_LVL, "dev-ip: could not set the vpn ip %s\n", argv[1]);
		return CMD_ERROR;
	}

	return CMD_SUCCESS;
}

static int dns_limit_cmd(int argc, char **argv)
{
	int val = 0;
	ptable_t *new_ptable = NULL;

	if (argc != 2)
	{
		log_msg(ERROR_LVL, "dns-limit: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	errno = 0; 
	val = (int) strtol(argv[1], NULL, 10);

	if ((errno == ERANGE && (val > 16 || val < 1))
		|| (errno != 0 && val == 0))
	{
		log_msg(ERROR_LVL, "dns-limit: wrong number, should be between 1-16.\n");
		return CMD_ERROR;
	}

	if (val == mdevice_g->dns_reqs->length)
		return CMD_SUCCESS;

	new_ptable = pt_create(PTABLE_FIXED | PTABLE_SEARCH, val, dnsreq_cmp);

	if (new_ptable == NULL)
	{
		log_msg(ERROR_LVL, "dns-limit: could not reset dns request limit, keeping default.\n");
		return CMD_ERROR;
	}

	pt_destroy(mdevice_g->dns_reqs);

	mdevice_g->dns_reqs = new_ptable;

	return CMD_SUCCESS;
}

static int dns_priority_cmd(int argc, char **argv)
{
	if (argc != 2)
	{
		log_msg(ERROR_LVL, "dns-priority: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	if (strncmp("ipv4", argv[1], 5) == 0)
		mdevice_g->prio_ipv4 = true;
	else if (strncmp("ipv6", argv[1], 5) == 0)
		mdevice_g->prio_ipv4 = false;
	else
	{
		log_msg(ERROR_LVL, "dns-priority: priority \"%s\" unknown\n", argv[1]);
		return CMD_ERROR;
	}

	return CMD_SUCCESS;
}

static int group_cmd(int argc, char **argv)
{
	struct group *grp = NULL;

	if (argc != 2)
	{
		log_msg(ERROR_LVL, "group: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	if ((grp = getgrnam (argv[1])) == NULL)
	{
		log_msg(ERROR_LVL, "group: no group named %s found\n", argv[1]);
		return CMD_ERROR;
	}

	if (setgid(grp->gr_gid) < 0)
	{
		log_msg(ERROR_LVL, "group: could not switch group\n");
		return CMD_ERROR;
	}

	return CMD_SUCCESS;
}

static int help_cmd(int argc, char **argv)
{
	command_t *cmd_run = NULL;
	int idx;
	int lst_idx;

	if (argc > 2)
	{
		log_msg(ERROR_LVL, "help: can take one argument (and one only)\n");
		return CMD_ERROR;
	}

	if (argc == 1)
	{
		log_msg(INFO_LVL, "Command list:\n");

		for (idx = 0, lst_idx = 1; idx < command_set->count; idx++, lst_idx++)
		{
			log_msg(INFO_LVL, "%s  ", ((command_t **) command_set->ptable)[idx]->name);

			if (lst_idx == 4)
			{
				log_msg(INFO_LVL, "\n");
				lst_idx = 0;
			}
		}
	}
	else
	{
		cmd_run = retrieve_command(argv[1], NULL);

		if (cmd_run  == NULL)
		{
			log_msg(ERROR_LVL, "help: no such command %s\n", argv[1]);
			return CMD_ERROR;
		}

		log_msg(INFO_LVL, "%s\n", cmd_run->help);

		switch (cmd_run->level)
		{
		case CMD_EXTERNAL:
			log_msg(INFO_LVL, "%s command can be used at configuration and runtime levels.\n", cmd_run->name);
			break;
		case CMD_INTERNAL:
			log_msg(INFO_LVL, "%s command can be used only at configuration level.\n", cmd_run->name);
			break;
		default:
			break;
		}
	}

	return CMD_SUCCESS;
}

static int limit_cmd(int argc, char **argv)
{
	int val = 0;

	if (argc != 2)
	{
		log_msg(ERROR_LVL, "limit: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	errno = 0; 
	val = (int) strtol(argv[1], NULL, 10);

	if ((errno == ERANGE && (val == INT_MAX || val == INT_MIN))
		|| (errno != 0 && val == 0)) 
	{
		log_msg(ERROR_LVL, "limit: wrong number\n");
		return CMD_ERROR;
	}

	mdevice_g->mnodes_limit = val;

	return CMD_SUCCESS;
}

static int load_cmd(int argc, char **argv)
{
	if (argc != 2)
	{
		log_msg(ERROR_LVL, "load: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	return load_plugin(argv[1]);
}

static int log_file_cmd(int argc, char **argv)
{
	if (argc > 2)
	{
		log_msg(ERROR_LVL, "log-file: can take one argument at most\n");
		return CMD_ERROR;
	}

	if (argc == 2)
		return init_logger(argv[1]);

	return init_logger(NULL);	
}

static int log_level_cmd(int argc, char **argv)
{
	log_level_t lvl;

	if (argc != 2)
	{
		log_msg(ERROR_LVL, "log-level: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	lvl = atoi(argv[1]);

	if (lvl >= MAX_LOG_LVL)
		log_msg(WARNING_LVL, "log-level %s unknown, using info level.\n", argv[1]);

	init_log_level(lvl);

	return CMD_SUCCESS;
}

static int persist_cmd(int argc, char **argv)
{
	bool persist = false;

	if (argc != 2)
	{
		log_msg(ERROR_LVL, "persist: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	if (atoi(argv[1]) > 0)
		persist = true;

	if (set_device_persist(mdevice_g, persist) < 0)
	{
		log_msg(ERROR_LVL, "persist: could not change device's persistance.\n");
		return CMD_ERROR;
	}

	return CMD_SUCCESS;
}

static int plugin_dir_cmd(int argc, char **argv)
{
	if (argc != 2)
	{
		log_msg(ERROR_LVL, "plugin-dir: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	strncpy(plugin_dir, argv[1], 512);

	return CMD_SUCCESS;
}

static int server_cmd(int argc, char **argv)
{
	mserver_t *new_mserver = NULL;
	int ret;

	if (argc != 2)
	{
		log_msg(ERROR_LVL, "server: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	if ((new_mserver = init_mserver()) == NULL)
	{
		log_msg(ERROR_LVL, "server: could not get memory for a new server\n");
		return CMD_ERROR;
	}

	ret = parse_connector_line(argv[1], &new_mserver->connector_id, &new_mserver->cipher_id,
							   new_mserver->host, &new_mserver->port);
	if (ret < 0)
	{
		log_msg(ERROR_LVL, "server line badly written: %s - ret code %d\n", argv[1], ret);
		free_mserver(new_mserver);
		return CMD_ERROR;
	}

	if (hostname_lookup(mdevice_g, new_mserver->host, mserver_dns_resolution, new_mserver) < 0)
	{
		log_msg(ERROR_LVL, "Could not proceed in hostname resolution\n");
		free_mserver(new_mserver);
		return CMD_ERROR;
	}

	return CMD_SUCCESS;
}

static int timeout_cmd(int argc, char **argv)
{
	int val = 0;

	if (argc != 2)
	{
		log_msg(ERROR_LVL, "timeout: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	errno = 0; 
	val = (int) strtol(argv[1], NULL, 10);

	if ((errno == ERANGE && (val > 600 || val < 10))
		|| (errno != 0 && val == 0))
	{
		log_msg(ERROR_LVL, "timeout: wrong number, should be between 10-600\n");
		return CMD_ERROR;
	}

	mdevice_g->timeout.tv_sec = val;

	return CMD_SUCCESS;
}

static int user_cmd(int argc, char **argv)
{
	struct passwd *pwd = NULL;

	if (argc != 2)
	{
		log_msg(ERROR_LVL, "user: needs one argument (and one only)\n");
		return CMD_ERROR;
	}

	if ((pwd = getpwnam(argv[1])) == NULL)
	{
		log_msg(ERROR_LVL, "user: no user named %s found\n", argv[1]);
		return CMD_ERROR;
	}

	if (setuid(pwd->pw_uid) < 0)
	{
		log_msg(ERROR_LVL, "user: could not switch user\n");
		return CMD_ERROR;
	}

	return CMD_SUCCESS;
}
