/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <dns.h>

static void ares_events(int fd, short event, void *arg);
static int dns_resolution(mdns_req_t *req, int status, struct hostent *hostent, char *host, struct sockaddr *addr);
static void remove_dns_req(mdns_req_t *req);

extern mdevice_t *mdevice_g;

int dnsreq_cmp(const void *r1, const void*r2)
{
	mdns_req_t * req1 = (mdns_req_t *) r1;
	mdns_req_t * req2 = (mdns_req_t *) r2;

	if (req1->sock < req2->sock)
		return -1;

	if (req1->sock > req2->sock)
		return 1;

	return 0;
}

int hostname_lookup(mdevice_t *mdevice, char *host, dns_callback_f resolv_cb, void *arg)
{
	struct in_addr addr4;
	struct in6_addr addr6;
	int addr_family = AF_UNSPEC;
	int sock_bits = 0;
	int ares_socks[VPZONE_MAX_DNS_REQ];
	int idx;
	mdns_req_t *new_req = NULL;
	mdns_req_t search_req;
	struct timeval *to;
	struct timeval timeout;

	/*Can we do dns req?*/
	if (mdevice->dns_reqs->count == mdevice->dns_reqs->length)
		return -1;

	new_req = calloc(1, sizeof(mdns_req_t));
	if (new_req == NULL)
	{
		ERR("hostname_lookup()");
		return -1;
	}

	/*Here we should also check the ip layer prio: ip4 or ip6?*/
	if (inet_pton(AF_INET, host, &addr4) == 1)
		addr_family = AF_INET;
	else if (inet_pton(AF_INET6, host, &addr6) == 1)
		addr_family = AF_INET6;
	else
	{
		if (mdevice->prio_ipv4)
			addr_family = AF_INET;
		else
			addr_family = AF_INET6;
	}

	new_req->mthing = arg;

	if ((pt_insert(mdevice->dns_reqs, new_req)) < 0)
	{
		log_msg(ERROR_LVL, "hostname_lookup(): could not insert new request\n");
		free(new_req);
		return -1;
	}

	ares_gethostbyname(mdevice->dns_req_channel, host, addr_family, resolv_cb, new_req);

	/*We need to retrieve query socket (it might not exist: after a direct result)*/
	memset(ares_socks, 0, VPZONE_MAX_DNS_REQ*(sizeof(int)));
	memset(&search_req, 0, sizeof(mdns_req_t));

	sock_bits = ares_getsock(mdevice->dns_req_channel, ares_socks, VPZONE_MAX_DNS_REQ);
	for (idx = 0; idx < VPZONE_MAX_DNS_REQ; idx++)
	{
		if (ares_socks[idx] == 0)
			continue;

		search_req.sock = ares_socks[idx];
		if (pt_search(mdevice->dns_reqs, &search_req) != NULL)
			continue;

		break;
	}

	/*Should happen only if request was directly proceeded*/
	if (idx >= VPZONE_MAX_DNS_REQ)
	{
		DEBUG_MSG("hostname_lookup(): request was processed immediately(?)\n");
		return 0;
	}

	DEBUG_MSG("idx = %d, socket: %d\n", idx, ares_socks[idx]);
	new_req->sock = ares_socks[idx];

	event_set(&new_req->ev_dns, new_req->sock, EV_TIMEOUT | EV_PERSIST | EV_READ | EV_WRITE, 
			  ares_events, &mdevice->dns_req_channel);

	timerclear(&timeout);
	to = ares_timeout(mdevice->dns_req_channel, NULL, &timeout);

	event_add(&new_req->ev_dns, to);

	return 0;
}

void mnode_dns_resolution(void *arg, int status, int timeouts, struct hostent *hostent)
{
	mdns_req_t *req = (mdns_req_t *) arg;
	mnode_t *new_mnode = (mnode_t *) req->mthing;
	(void) timeouts;

	if (dns_resolution(req, status, hostent, new_mnode->host, &new_mnode->address) < 0)
	{	
		free_mnode(new_mnode);
		return;
	}
	
	if (connect_mnode(new_mnode) < 0)
	{
		log_msg(ERROR_LVL, "Could no connect mnode to %s:%u\n", new_mnode->host, new_mnode->port);
		free_mnode(new_mnode);
	}

	remove_dns_req(req);
}

void mserver_dns_resolution(void *arg, int status, int timeouts, struct hostent *hostent)
{
	mdns_req_t *req = (mdns_req_t *) arg;
	mserver_t *new_mserver = (mserver_t *) req->mthing;
	(void) timeouts;

	if (dns_resolution(req, status, hostent, new_mserver->host, &new_mserver->address) < 0)
	{
		free_mserver(new_mserver);
		return;
	}
	
	if (setup_mserver(mdevice_g, new_mserver) < 0)
	{
		log_msg(ERROR_LVL, "Could no setup mserver on %s:%u\n", new_mserver->host, new_mserver->port);
		free_mserver(new_mserver);
	}

	remove_dns_req(req);
}


static void ares_events(int fd, short event, void *arg)
{
	ares_channel channel = *((ares_channel *)arg);
	ares_socket_t r = ARES_SOCKET_BAD;
	ares_socket_t w = ARES_SOCKET_BAD;

	if (event & EV_READ)
		r = fd;

	if (event & EV_WRITE)
		w = fd;

	if (event & EV_TIMEOUT)
		DEBUG_MSG("DNS request timeout\n");

	ares_process_fd(channel, r, w);
}

static int dns_resolution(mdns_req_t *req, int status, struct hostent *hostent, char *host, struct sockaddr *addr)
{
	struct in_addr *addr4;
	struct in6_addr *addr6;

	if ((status != ARES_SUCCESS) || (hostent == NULL))
	{
		log_msg(ERROR_LVL, "Could not resolve hostname: %s\n", host);
		remove_dns_req(req);
		return -1;
	}

	/*Let's consider first result only for now, later on: 
	  check the addr content and loop on result to proceed with other addresses.*/
	switch (hostent[0].h_addrtype)
	{
	case AF_INET:
		addr4 = &(((struct sockaddr_in *) addr)->sin_addr);
		memcpy(addr4, hostent[0].h_addr_list[0], sizeof(struct in_addr));
		addr->sa_family = AF_INET;
		break;
	case AF_INET6:
		addr6 = &(((struct sockaddr_in6 *) &addr)->sin6_addr);
		memcpy(addr6, hostent[0].h_addr_list[0], sizeof(struct in6_addr));
		addr->sa_family = AF_INET6;
		break;
	default:
		remove_dns_req(req);
		return -1;
	}

	return 0;
}

static void remove_dns_req(mdns_req_t *req)
{
	int idx;

	for (idx = 0; idx < mdevice_g->dns_reqs->count; idx++)
		if ((mdns_req_t *) mdevice_g->dns_reqs->ptable[idx] == req)
			break;

	pt_remove(mdevice_g->dns_reqs, idx);

	event_del(&req->ev_dns);
	free(req);
}
