/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <hash.h>

void *h_create(uint32_t size, hash_f hash_compute_function, keycmp_f key_cmp_function)
{
	hashmap_t *h_ret = NULL;

	h_ret = (hashmap_t *) calloc(1, sizeof(hashmap_t));
	if (h_ret == NULL)
		return NULL;

	h_ret->length = 1u;
	while (h_ret->length < size)
		h_ret->length = h_ret->length << 1;

	h_ret->entries = (hentry_t **) calloc(h_ret->length, sizeof(hentry_t *));
	if (h_ret->entries == NULL)
	{
		free(h_ret);
		return NULL;
	}

	h_ret->hash_compute = hash_compute_function;
	h_ret->key_compare = key_cmp_function;

	return (void *) h_ret;
}

int h_insert(void *hm, void *key, void *value)
{
	hentry_t *he;
	uint32_t index;
	hashmap_t *hmap = (hashmap_t *) hm;

	he = (hentry_t *) calloc(1, sizeof(hentry_t));
	if (he == NULL)
		return -1;

	index = HASH_INDEX(hmap->hash_compute(key), hmap->length);

	he->key = key;
	he->value = value;
	he->next = hmap->entries[index];

	hmap->entries[index] = he;
	hmap->nb_entries++;

	return 0;
}

void *h_search(void *hm, void *key)
{
	hentry_t *he = NULL;
	hashmap_t *hmap = (hashmap_t *) hm;
	uint32_t index;

	if (key == NULL)
		return NULL;

	index= HASH_INDEX(hmap->hash_compute(key), hmap->length);

	he = hmap->entries[index];
	
	while (he != NULL)
	{
		if (hmap->key_compare(key, he->key) == 0)
			return he->value;

		he = he->next;
	}

	return NULL;
}

void *h_remove(void *hm, void *key, bool free_key)
{
	hentry_t *cur_he = NULL;
	hentry_t *prev_he = NULL;
	uint32_t index;
	void *value = NULL;
	hashmap_t *hmap = (hashmap_t *) hm;

	if (key == NULL)
		return NULL;

	index = HASH_INDEX(hmap->hash_compute(key), 
					   hmap->length);

	cur_he = hmap->entries[index];

	while (cur_he != NULL)
	{
		if (hmap->key_compare(key, cur_he->key) == 0)
		{
			if (prev_he != NULL)
				prev_he->next = cur_he->next;
			else
				hmap->entries[index] = cur_he->next;

			/*If we are removing current iterator entry...*/
			if ((hmap->iterator.cur_entry == cur_he) && (hmap->iterator.index == index))
			{
				if (prev_he != NULL)
					hmap->iterator.parent = prev_he;
				else
					hmap->iterator.parent = hmap->entries[index];

				hmap->iterator.cur_entry = NULL;
			}

			value = cur_he->value;
			if (free_key)
				free(cur_he->key);
			hmap->nb_entries--;
			free(cur_he);

			return value;
		}

		prev_he = cur_he;
		cur_he = cur_he->next;
	}

	return NULL;
}

int h_destroy(void *hm)
{
	hentry_t *nxt_he = NULL;
	hentry_t *rem_he = NULL;
	uint32_t index;
	hashmap_t *hmap = (hashmap_t *) hm;

	if (hm == NULL)
		return 0;

	for (index = 0; index < hmap->length; index++)
	{
		nxt_he = hmap->entries[index];
		while (nxt_he != NULL)
		{
			rem_he = nxt_he;
			nxt_he = nxt_he->next;
			free(rem_he);
		}
	}

	free(hmap->entries);
	free(hmap);

	return 0;
}

uint32_t h_count(void *hm)
{
	return ((hashmap_t *)hm)->nb_entries;
}

int h_iter_begin(void *hm)
{
	hashmap_t *hmap = (hashmap_t *) hm;

	if ((hmap->iterator.cur_entry != NULL) || 
		(hmap->iterator.parent != NULL))
		return -1;

	return 0;
}

void h_iter_end(void *hm)
{
	hashmap_t *hmap = (hashmap_t *) hm;

	hmap->iterator.cur_entry = NULL;
	hmap->iterator.parent = NULL;
	hmap->iterator.index = (uint32_t) -1;
}

bool h_iter_next(void *hm)
{
	hashmap_t *hmap = (hashmap_t *) hm;

	if (hmap->iterator.index == hmap->length)
		return false;

	if (hmap->iterator.parent != NULL)
	{
		/*cur_entry has been removed*/
		if (hmap->iterator.parent->next != NULL)
		{
			hmap->iterator.cur_entry = hmap->iterator.parent->next;
			hmap->iterator.parent = NULL;
			return true;
		}

		hmap->iterator.parent = NULL;
	}
	else
	{
		if ((hmap->iterator.cur_entry != NULL) && 
			(hmap->iterator.cur_entry->next != NULL))
		{
			hmap->iterator.cur_entry = hmap->iterator.cur_entry->next;
			return true;
		}
	}

	while ((hmap->iterator.index < hmap->length) || (hmap->iterator.index == (uint32_t) -1))
	{
		/*We just proceed with indexes*/
		hmap->iterator.index++;

		if (hmap->entries[hmap->iterator.index] != NULL)
		{
			hmap->iterator.cur_entry = hmap->entries[hmap->iterator.index];
			return true;
		}
	}

	return false;
}

void *h_iter_getkey(void *hm)
{
	hashmap_t *hmap = (hashmap_t *) hm;

	if (hmap->iterator.cur_entry == NULL)
		return NULL;

	return hmap->iterator.cur_entry->key;
}

void *h_iter_getvalue(void *hm)
{
	hashmap_t *hmap = (hashmap_t *) hm;

	if (hmap->iterator.cur_entry == NULL)
		return NULL;

	return hmap->iterator.cur_entry->value;
}
