/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iface.h>

int open_tun_device(mdevice_t *mdevice)
{
	mdevice->tun_device = open("/dev/net/tun", O_RDWR | O_NONBLOCK);
	if (mdevice->tun_device < 0)
	{
		ERR("open_tun_device()");
		return -1;
	}

	return 0;
}

int close_tun_device(mdevice_t *mdevice)
{
	if (close(mdevice->tun_device) < 0)
	{
		ERR("close_tun_device()");
		return -1;
	}

	return 0;
}

int set_device_name(mdevice_t *mdevice)
{
	struct ifreq ifr_tun;

	memset(&ifr_tun, 0, sizeof(struct ifreq));

	ifr_tun.ifr_flags = IFF_TUN | IFF_NO_PI;

	if (*mdevice->dev_name)
		strncpy(ifr_tun.ifr_name, mdevice->dev_name, IF_NAMESIZE);

    if(ioctl(mdevice->tun_device, TUNSETIFF, &ifr_tun) < 0)
	{
		ERR("set_device_name()");
		return -1;
	}

	strncpy(mdevice->dev_name, ifr_tun.ifr_name, IF_NAMESIZE);

	return 0;
}

int set_device_persist(mdevice_t *mdevice, bool persist)
{
	if(ioctl(mdevice->tun_device, TUNSETPERSIST, (persist ? 1 : 0)) < 0)
	{
		ERR("set_device_persist()");
		return -1;
	}

	return 0;
}

int set_device_address(mdevice_t *mdevice, struct in6_addr address, short prefix_len)
{
	int sock_tun;
	struct ifreq ifr_tun;
	struct in6_ifreq in6_tun;

	if ((sock_tun = socket(AF_INET6, SOCK_DGRAM, 0)) < 0)
	{
		ERR("set_device_address()");
		return -1;
	}

	memset(&ifr_tun, 0, sizeof(struct ifreq));
	strncpy (ifr_tun.ifr_name, mdevice->dev_name, IF_NAMESIZE);

	if (ioctl(sock_tun, SIOGIFINDEX, &ifr_tun) < 0)
	{
		ERR("set_device_address()");
		close(sock_tun);
		return -1;
	}

	memset(&in6_tun, 0, sizeof(struct in6_ifreq));
	in6_tun.ifr6_ifindex = ifr_tun.ifr_ifindex;
	in6_tun.ifr6_prefixlen = prefix_len;
	memcpy(&in6_tun.ifr6_addr, &address, sizeof(struct in6_addr));

	if(ioctl(sock_tun, SIOCSIFADDR, &in6_tun) < 0)
	{
		if (errno != EEXIST)
		{
			ERR("set_device_address()");
			close(sock_tun);
			return -1;
		}
	}

	close(sock_tun);

	if (IN6_IS_ADDR_LINKLOCAL(&address))
	{
		DEBUG_MSG("Is a LINK LOCAL address\n");
		memcpy(&mdevice->vpn_uip, &address, sizeof(struct in6_addr));
	}
	else
	{
		DEBUG_MSG("Is NOT a LINK LOCAL address\n");
		memcpy(&mdevice->vpn_ip, &address, sizeof(struct in6_addr));
	}

	return 0;
}

int set_device_up(mdevice_t *mdevice, bool up)
{
	int sock_tun;
	struct ifreq ifr_tun;

	if ((sock_tun = socket(AF_INET6, SOCK_DGRAM, 0)) < 0)
	{
		ERR("set_device_up()");
		return -1;
	}

	memset(&ifr_tun, 0, sizeof(struct ifreq));
	strncpy (ifr_tun.ifr_name, mdevice->dev_name, IF_NAMESIZE);

	if (ioctl(sock_tun, SIOCGIFFLAGS, &ifr_tun) < 0)
	{
		ERR("set_device_up()");
		close(sock_tun);
		return -1;
	}

	if ((up) && !((IFF_UP | IFF_RUNNING) & ifr_tun.ifr_flags))
		ifr_tun.ifr_flags |= IFF_UP | IFF_RUNNING;

	if (!(up) && ((IFF_UP | IFF_RUNNING) & ifr_tun.ifr_flags))
		ifr_tun.ifr_flags ^= IFF_RUNNING;

	if (ioctl(sock_tun, SIOCSIFFLAGS, &ifr_tun) < 0)
	{
		ERR("set_device_up()");
		close(sock_tun);
		return -1;
	}

	close(sock_tun);

	return 0;
}
