/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <io.h>

extern mdevice_t *mdevice_g;

void manage_packet_df(mnode_t *mnode)
{
	int p_len = 0;
	uint8_t *pkt_hdr;

	while (true)
	{
		if (mnode->r_bytes <= TUN_PKT_LEN)
			break;

		p_len = 0;

		pkt_hdr = mnode->r_buffer;

		DBG_PACKET_HDR("Managing packet", pkt_hdr);

		switch (IP6_VER(pkt_hdr))
		{
		case IPV6_VERSION:
			if (IN6_ARE_ADDR_EQUAL(IP6_DST(pkt_hdr), &(mdevice_g->vpn_uip)) ||
				IN6_ARE_ADDR_EQUAL(IP6_DST(pkt_hdr), &(mdevice_g->vpn_ip)) ||
				IN6_IS_ADDR_MULTICAST(IP6_DST(pkt_hdr)))
			{
				p_len = IP6_LEN(pkt_hdr) + TUN_PKT_LEN;
				if (p_len > mnode->r_bytes)
					return;

				memcpy(mdevice_g->w_buffer+mdevice_g->w_bytes, mnode->r_buffer, p_len);
				mdevice_g->w_bytes += p_len;

				write_tun(mdevice_g->tun_device, EV_WRITE, mdevice_g);
			}
			else
				p_len = route_packet(mnode->r_buffer, mnode->r_bytes, pkt_hdr);
			break;
		default:
			manage_internal_packet(p_len, mdevice_g, mnode);
			break;
		}

		if (p_len <= 0)
			break;

		if (mnode->r_bytes > p_len)
			memmove(mnode->r_buffer, mnode->r_buffer + p_len, mnode->r_bytes - p_len);

		mnode->r_bytes -= p_len;
		if (mnode->r_bytes < 0)
			mnode->r_bytes = 0;
	}
}

int route_packet_df(uint8_t *buffer, int buff_len, void *arg)
{
	mnode_t *mnode_found = NULL;
	int p_len = 0;
	uint8_t *pkt_hdr = (uint8_t *) arg;

	mnode_found = h_search(mdevice_g->ip_mnodes, IP6_DST(pkt_hdr));
	if (mnode_found == NULL)
	{
		select_route(mnode_found, (struct in6_addr *) IP6_DST(pkt_hdr))
		if (mnode_found == NULL)
			return -1;
	}

	p_len = IP6_LEN(pkt_hdr) + TUN_PKT_LEN;
	if (p_len > buff_len)
		return -1;

	if ((mnode_found->w_bytes+p_len) > mdevice_g->buffer_size)
		return -1;

	memcpy(mnode_found->w_buffer + mnode_found->w_bytes, buffer, p_len);
	mnode_found->w_bytes += p_len;

	event_add(&mnode_found->write_event, NULL);

	return p_len;
}

void read_tun_df(int fd, short event, void *arg)
{
	int in_bytes = 0;
	int p_len = 0;
	uint8_t trans_buff[TUN_DEV_MTU];
	uint8_t *pkt_hdr;
	mnode_t *mnode = NULL;
	mdevice_t *mdevice = (mdevice_t *)arg;

	/*Make compiler happy*/
	if ((fd != mdevice->tun_device) || (event != EV_READ))
		return;
	/***/

	memset(trans_buff, 0, TUN_DEV_MTU);

	in_bytes = read(mdevice->tun_device, trans_buff, TUN_DEV_MTU);
	if (in_bytes < 0)
	{
		if ((errno == EAGAIN) || (errno == EINTR))
			return;
		/*THIS IS FATAL!*/
		ERR("read_tun_df()");
		log_msg(FATAL_LVL, "TUN device fatal error - exiting.\n");
		exit(EXIT_FAILURE);
	}

	if (in_bytes < TUN_PKT_LEN)
		return;

	pkt_hdr = trans_buff;
	if (IP6_VER(pkt_hdr) != IPV6_VERSION)
		return;

	DBG_PACKET_HDR("Read TUN", pkt_hdr);

	if (IN6_IS_ADDR_MULTICAST(IP6_DST(pkt_hdr)))
	{
		if (h_count(mdevice->ip_mnodes) == 0)
			return;

		h_iter_begin(mdevice_g->ip_mnodes);

		p_len = IP6_LEN(pkt_hdr) + TUN_PKT_LEN;

		while(h_iter_next(mdevice_g->ip_mnodes))
		{
			if (is_ipv6_unicast(h_iter_getkey(mdevice_g->ip_mnodes), NULL) == 0)
			{
				mnode = h_iter_getvalue(mdevice_g->ip_mnodes);
				memcpy(mnode->w_buffer + mnode->w_bytes, trans_buff, p_len);
				mnode->w_bytes += p_len;

				event_add(&mnode->write_event, NULL);
			}
		}

		h_iter_end(mdevice_g->ip_mnodes);
	}
	else
		p_len = route_packet(trans_buff, in_bytes, pkt_hdr);
}

void write_tun_df(int fd, short event, void *arg)
{
	int out_bytes = 0;
	mdevice_t *mdevice = (mdevice_t *) arg;

	/*Make compiler happy*/
	if ((fd != mdevice->tun_device) || (event != EV_WRITE))
		return;
	/***/

	DEBUG_MSG("TUN writing\n");

	out_bytes = write(mdevice->tun_device, mdevice->w_buffer, mdevice->w_bytes);
	if (out_bytes <= 0)
	{
		if ((errno == EAGAIN) || (errno == EINTR))
		{
			event_add(&mdevice->write_event, NULL);
			return;
		}
		/*THIS IS FATAL!*/
		ERR("write_tun_df()");
		log_msg(FATAL_LVL, "TUN device fatal error - exiting.\n");
		exit(EXIT_FAILURE);
	}

	mdevice->w_bytes -= out_bytes;
	if (mdevice->w_bytes > 0)
	{
		memmove(mdevice->w_buffer, mdevice->w_buffer + out_bytes, mdevice->w_bytes);
		event_add(&mdevice->write_event, NULL);
	}
	else
	{
		mdevice->w_bytes = 0;
		event_del(&mdevice->write_event);
	}
}

void remove_mnode_df(mdevice_t *mdevice, mnode_t *mnode)
{
	if (mnode == NULL)
		return;

	pre_disconnect(mnode);

	if (mnode->is_server)
	{
		deinit_mnode_server_cipher(mnode);
	}
	else
	{
		deinit_mnode_client_cipher(mnode);
	}

	disconnect_mnode(mnode);

	event_del(&mnode->read_event);
	event_del(&mnode->write_event);

	h_remove(mdevice->ip_mnodes, &(mnode->vpn_ip), false);
	h_remove(mdevice->ip_mnodes, &(mnode->vpn_uip), false);

	log_msg(INFO_LVL, "Removed mnode\n");

	free_mnode(mnode);
}

void remove_mserver_df(mdevice_t *mdevice, mserver_t *mserver)
{
	mserver_t *prev_mserver = NULL;
	mserver_t *cur_mserver = NULL;
	mnode_t *mnode = NULL;

	close_mserver(mdevice, mserver);

	if (h_count(mdevice->ip_mnodes) > 0)
	{
		h_iter_begin(mdevice_g->ip_mnodes);

		while(h_iter_next(mdevice_g->ip_mnodes))
		{
			if (is_ipv6_unicast(h_iter_getkey(mdevice_g->ip_mnodes), NULL) == 0)
			{
				mnode = h_iter_getvalue(mdevice_g->ip_mnodes);
				if ((mnode->is_server) && (mnode->connector_id == mserver->connector_id))
					remove_mnode(mdevice_g, mnode);
			}
		}

		h_iter_end(mdevice_g->ip_mnodes);
	}

	event_del(&mserver->server_event);

	cur_mserver = mdevice->mservers;
	while ((cur_mserver != mserver) && (cur_mserver != NULL))
	{
		prev_mserver = cur_mserver;
		cur_mserver = cur_mserver->next;
	}

	if (cur_mserver == NULL)
		cur_mserver = mserver;
	else
	{
		if (prev_mserver != NULL)
			prev_mserver->next = cur_mserver->next;
		else
			mdevice->mservers = cur_mserver->next;
	}

	log_msg(INFO_LVL, "Removed mserver\n");

	free_mserver(cur_mserver);
}

void remove_mdevice_df(void)
{
	event_del(&mdevice_g->read_event);
	event_del(&mdevice_g->write_event);

	if (mdevice_g->base_event != NULL)
		event_base_free(mdevice_g->base_event);

	h_destroy(mdevice_g->ip_mnodes);

	close_tun_device(mdevice_g);

	log_msg(INFO_LVL, "Removed mdevice\n");

	free_mdevice(mdevice_g);
}

void setup_io(mdevice_t *mdevice)
{
	struct timeval timeout;
	int status;

	DEBUG_MSG("Init LibEvent\n");
	mdevice->base_event = event_init();
	if (mdevice->base_event == NULL)
	{
		log_msg(FATAL_LVL, "Libevent could not initialize!\n");
		exit(EXIT_FAILURE);
	}

	/*Enabling async DNS resolution*/
#ifdef CARES_HAVE_ARES_LIBRARY_INIT
	status = ares_library_init(ARES_LIB_INIT_ALL);
	if (status != ARES_SUCCESS)
	{
		log_msg(FATAL_LVL, "Could not set up c-ares library: %s\n", ares_strerror(status));
		exit(EXIT_FAILURE);
	}
#endif

	status = ares_init(&(mdevice->dns_req_channel));
	if (status != ARES_SUCCESS)
    {
		log_msg(FATAL_LVL, "Could not set up c-ares : %s\n", ares_strerror(status));
		exit(EXIT_FAILURE);
    }

	timerclear(&timeout);
	ares_timeout(mdevice->dns_req_channel, NULL, &timeout);
}

void poll_io(mdevice_t *mdevice)
{
	DEBUG_MSG("Adding TUN device events\n");
	event_set(&mdevice->read_event, mdevice->tun_device, 
			  EV_READ|EV_PERSIST, read_tun, mdevice);

	event_set(&mdevice->write_event, mdevice->tun_device, 
			  EV_WRITE|EV_PERSIST, write_tun, mdevice);

	event_add(&mdevice->read_event, NULL);

	DEBUG_MSG("Dispatching...\n");
	if (event_base_dispatch(mdevice->base_event) < 0)
		log_msg(FATAL_LVL, "LibEvent event base loop exited abnormally\n");
}
