/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <plugin.h>

static plugin_t *init_plugin(void);
static void *get_symbol(const char *symbol_str, plugin_t *plugin);
static void _unload_plugin(plugin_t *plugin);
static int check_dependencies(const char *deps);

char plugin_dir[512];

ptable_t *mnode_plugins = NULL;
ptable_t *connectors = NULL;
ptable_t *ciphers = NULL;
ptable_t *authenticators = NULL;

in_plugin_handlers_t *iph = NULL;

static bool is_hooked[HDL_MAX_HANDLERS];

extern const handler_t default_ih[HDL_MAX_HANDLERS];
extern mdevice_t *mdevice_g;

int load_plugin(char *plugin_name)
{
	char plugin_file_name[512];
	char *msg;
	plugin_t *new_plugin = NULL;
	int plugin_id = -1;
	int idx;

	connector_t *conn = NULL;
	get_connectors_f get_connectors;
	cipher_t *ciph = NULL;
	get_ciphers_f get_ciphers;
	authenticator_t *auth = NULL;
	get_authenticators_f get_authenticators;

	if (get_plugin_by_name(plugin_name) != NULL)
	{
		log_msg(INFO_LVL, "Plugin %s already loaded\n", plugin_name);
		return 0;
	}

	DEBUG_MSG("Loading plugin: %s\n", plugin_name);

	if ((new_plugin = init_plugin()) == NULL)
		return -1;

	strncpy(new_plugin->name, plugin_name, PLUGIN_NAME_MAX_LENGTH);

	snprintf(plugin_file_name, 512, "%s/%s.so", plugin_dir, new_plugin->name);

	if ((new_plugin->handle = dlopen(plugin_file_name, RTLD_NOW)) == NULL)
	{
		log_msg(ERROR_LVL, "Cannot load plugin %s: %s\n", new_plugin->name, dlerror());
		_unload_plugin(new_plugin);
		return -1;
	}
	
	if ((new_plugin->unload = get_symbol("unload", new_plugin)) == NULL)
		return -1;

	if ((new_plugin->init = get_symbol("init", new_plugin)) == NULL)
		return -1;
	
	dlerror();
	new_plugin->dependencies = *(char **) dlsym(new_plugin->handle, "dependencies");
	if (((msg = dlerror()) != NULL) || (new_plugin->dependencies == NULL))
		log_msg(INFO_LVL, "Plugin %s has no dependencies\n", new_plugin->name);
	else
	{
		log_msg(INFO_LVL, "Plugin %s has dependencies - Checking...%s\n", new_plugin->name, new_plugin->dependencies);
		if (check_dependencies(new_plugin->dependencies) < 0)
		{
			_unload_plugin(new_plugin);
			return -1;
		}
	}

	plugin_id = pt_insert(mnode_plugins, (void *) new_plugin);
	if (plugin_id < 0)
	{
		_unload_plugin(new_plugin);
		return -1;
	}

	new_plugin->id = plugin_id;

	new_plugin->type = new_plugin->init(plugin_id, iph);
	if (new_plugin->type < 0)
	{
		DEBUG_MSG("Plugin init() failed\n");
		_unload_plugin(new_plugin);
		return -1;
	}

	if (new_plugin->type & PLUGIN_CONNECTOR)
	{
		get_connectors = get_symbol("get_connectors", new_plugin);
		if (get_connectors == NULL)
			return -1;

		if ((conn = get_connectors()) == NULL)
		{
			log_msg(ERROR_LVL, "Cannot get connectors from %s\n", new_plugin->name);
			_unload_plugin(new_plugin);
			return -1;
		}

		for (idx = 0; conn[idx].plugin_id != -1; idx++)
		{
			if (pt_insert(connectors, (void *) &(conn[idx])) < 0)
			{
				log_msg(ERROR_LVL, "Cannot insert connector %s\n", conn[idx].type);
				_unload_plugin(new_plugin);
				return -1;
			}
		}
	}

	if (new_plugin->type & PLUGIN_CIPHER)
	{
		get_ciphers = get_symbol("get_ciphers", new_plugin);
		if (get_ciphers == NULL)
			return -1;

		if ((ciph = get_ciphers()) == NULL)
		{
			log_msg(ERROR_LVL, "Cannot get ciphers from %s\n", new_plugin->name);
			_unload_plugin(new_plugin);
			return -1;
		}

		for (idx = 0; ciph[idx].plugin_id != -1; idx++)
		{
			if (pt_insert(ciphers, (void *) &(ciph[idx])) < 0)
			{
				log_msg(ERROR_LVL, "Cannot insert cipher %s\n", ciph[idx].type);
				_unload_plugin(new_plugin);
				return -1;
			}
		}
	}

	if (new_plugin->type & PLUGIN_AUTHENTICATOR)
	{
		get_authenticators = get_symbol("get_authenticators", new_plugin);
		if (get_authenticators == NULL)
			return -1;

		if ((auth = get_authenticators()) == NULL)
		{
			log_msg(ERROR_LVL, "Cannot get authenticators from %s\n", new_plugin->name);
			_unload_plugin(new_plugin);
			return -1;
		}

		for (idx = 0; auth[idx].plugin_id != -1; idx++)
		{
			if (pt_insert(authenticators, (void *) &(auth[idx])) < 0)
			{
				log_msg(ERROR_LVL, "Cannot insert authenticator %s\n", auth[idx].type);
				_unload_plugin(new_plugin);
				return -1;
			}
		}
	}

	return 0;
}

static void *get_symbol(const char *symbol_str, plugin_t *plugin)
{
	void *symbol = NULL;
	char *msg;

	dlerror();
	symbol = dlsym(plugin->handle, symbol_str);
	if ((msg = dlerror()) != NULL)
	{
		log_msg(WARNING_LVL, "Cannot get %s in %s - %s\n", symbol_str, plugin->name, msg);
		_unload_plugin(plugin);
		return NULL;
	}

	return symbol;
}

int unload_plugin(char *plugin_name)
{
	plugin_t *plugin = NULL;
	int idx = 0;

	if ((strlen(plugin_name) == 0) || (mnode_plugins == NULL))
		return 0;

	plugin = get_plugin_by_name(plugin_name);
	if (plugin == NULL)
		return -1;

	log_msg(INFO_LVL, "Unloading plugin: %s\n", plugin_name);
	
	if ((plugin->type & PLUGIN_CONNECTOR) && (connectors->count > 0))
	{
		for (idx = connectors->length-1; idx >= 0; idx--)
			if ((connectors->ptable[idx] != NULL) && 
				(plugin->id == ((connector_t *) connectors->ptable[idx])->plugin_id))
				pt_remove(connectors, idx);
	}

	if ((plugin->type & PLUGIN_CIPHER) && (ciphers->count > 0))
	{
		for (idx = ciphers->length-1; idx >= 0; idx--)
			if ((ciphers->ptable[idx] != NULL) && 
				(plugin->id == ((cipher_t *) ciphers->ptable[idx])->plugin_id))
				pt_remove(ciphers, idx);
	}

	if ((plugin->type & PLUGIN_AUTHENTICATOR) && (authenticators->count > 0))
	{
		for (idx = authenticators->length-1; idx >= 0; idx--)
			if ((authenticators->ptable[idx] != NULL) && 
				(plugin->id == ((authenticator_t *) authenticators->ptable[idx])->plugin_id))
				pt_remove(authenticators, idx);
	}

	_unload_plugin(plugin);

	return 0;
}

static void _unload_plugin(plugin_t *plugin)
{
	if (plugin->unload != NULL)
		if (plugin->unload() < 0)
			log_msg(WARNING_LVL, "Error in plugin unload in %s\n", plugin->name);

	if (plugin->handle != NULL)
		dlclose(plugin->handle);

	if (plugin->id > -1)
		pt_remove(mnode_plugins, plugin->id);

	free(plugin);
}

static plugin_t *init_plugin(void)
{
	plugin_t *new_plugin = NULL;

	new_plugin = (plugin_t *) calloc(1, sizeof(plugin_t));
	if(new_plugin == NULL)
	{
		ERR("init_plugin()");
		return NULL;
	}

	new_plugin->id = -1;

	return new_plugin;
}

static int check_dependencies(const char *deps)
{
	char *pos = NULL;
	size_t len = 0;
	char plugin_name[PLUGIN_NAME_MAX_LENGTH];

	while (((pos = strchr(deps, ' ')) != NULL) || 
		   ((len = strlen(deps)) > 0))
	{
		memset(plugin_name, '\0', PLUGIN_NAME_MAX_LENGTH);
		
		if (pos == NULL)
		{
			if (len < PLUGIN_NAME_MAX_LENGTH)
				strncpy(plugin_name, deps, len);
			else
				break;
		}
		else
			if (((pos - deps) < PLUGIN_NAME_MAX_LENGTH) && ((pos - deps) > 0))
				strncpy(plugin_name, deps, pos - deps);

		if (strlen(plugin_name) > 0)
		{
			if (load_plugin(plugin_name) < 0)
			{
				log_msg(ERROR_LVL, "Denpendency plugin %s lacks\n", plugin_name);
				return -1;
			}
		}

		if (pos != NULL)
			deps = pos+1;
		else 
			break;
	}

	return 0;
}

plugin_t *get_plugin_by_name(char *plugin_name)
{
	plugin_t *plugin = NULL;
	int idx;

	if ((strlen(plugin_name) == 0) || (mnode_plugins->count <= 0))
		return NULL;

	for (idx = 0; idx < mnode_plugins->length; idx++)
	{
		if (mnode_plugins->ptable[idx] == NULL)
			continue;

		if (strncmp(((plugin_t *) mnode_plugins->ptable[idx])->name, plugin_name, strlen(plugin_name)) == 0)
		{
			plugin = (plugin_t *) mnode_plugins->ptable[idx];
			break;
		}
	}

	return plugin;
}

int set_handler(handler_num_t handler_num, handler_t handler)
{	
	if (handler_num > HDL_MAX_HANDLERS)
		return HOOK_UNKNOWN;
	
	if (is_hooked[handler_num])
			return HOOK_ALREADY_SET;

	ih[handler_num] = handler;
	iph->handlers[handler_num] = ih[handler_num];

	is_hooked[handler_num] = true;

	return HOOK_SUCCESSFUL;
}

int unset_handler(handler_num_t handler_num)
{	
	if (handler_num > HDL_MAX_HANDLERS)
		return HOOK_UNKNOWN;
	
	if (!is_hooked[handler_num])
			return HOOK_SUCCESSFUL;

	ih[handler_num] = default_ih[handler_num];
	iph->handlers[handler_num] = ih[handler_num];

	is_hooked[handler_num] = false;

	return HOOK_SUCCESSFUL;
}

int add_callback(callback_num_t callback_num, handler_t callback)
{
	if (callback_num > CB_MAX_CALLBACKS)
		return HOOK_UNKNOWN;

	icb[callback_num] = insert_callback(icb[callback_num], callback);
	iph->callbacks[callback_num] = icb[callback_num];

	return HOOK_SUCCESSFUL;
}

int sub_callback(callback_num_t callback_num, handler_t callback)
{
	if (callback_num > CB_MAX_CALLBACKS)
		return HOOK_UNKNOWN;

	icb[callback_num] = remove_callback(icb[callback_num], callback);
	iph->callbacks[callback_num] = icb[callback_num];

	return HOOK_SUCCESSFUL;
}

callback_t *insert_callback(callback_t *callback_cb, handler_t callback)
{
	callback_t *cur_cb = NULL;
	callback_t *prev_cb = NULL;

	if (callback_cb != NULL)
	{
		for (cur_cb = prev_cb = callback_cb; 
			 cur_cb != NULL; 
			 prev_cb = cur_cb, cur_cb = cur_cb->next)
			if (cur_cb->callback == callback)
				return callback_cb;
	}

	cur_cb = (callback_t *) calloc(1, sizeof(callback_t));
	if (cur_cb == NULL)
	{
		ERR("insert_callback()");
		return callback_cb;
	}

	cur_cb->callback = callback;	

	if (callback_cb != NULL)
	{
		prev_cb->next = cur_cb;
		return callback_cb;
	}
	
	return cur_cb;
}

callback_t *remove_callback(callback_t *callback_cb, handler_t callback)
{
	callback_t *cur_cb = NULL;
	callback_t *prev_cb = NULL;

	if (callback_cb == NULL)
		return NULL;

	for (cur_cb = prev_cb = callback_cb; 
		 cur_cb != NULL && (cur_cb->callback != callback); 
		 prev_cb = cur_cb, cur_cb = cur_cb->next);

	if (cur_cb != NULL)
	{
		if (cur_cb != prev_cb)
			prev_cb->next = cur_cb->next;
		else
			callback_cb = cur_cb->next;
		
		free(cur_cb);
	}
	
	return callback_cb;
}

void setup_plugin_manager(void)
{
	iph = (in_plugin_handlers_t *) calloc(1, sizeof(in_plugin_handlers_t));
	if (iph == NULL)
	{
		ERR("setup_plugin_manager()\n");
		exit(EXIT_FAILURE);
	}

	strncpy(plugin_dir, PLUGIN_DEFAULT_DIR, 512);

	iph->log_msg_h = log_msg;
	iph->log_set_redirection_h = log_set_redirection;
	iph->log_unset_redirection_h = log_unset_redirection;
	iph->set_handler_h = set_handler;
	iph->unset_handler_h = unset_handler;
	iph->add_callback_h = add_callback;
	iph->sub_callback_h = sub_callback;
	iph->init_mnode_h = init_mnode;
	iph->free_mnode_h = free_mnode;
	iph->pt_create_h = pt_create;
	iph->pt_insert_h = pt_insert;
	iph->pt_search_h = pt_search;
	iph->pt_remove_h = pt_remove;
	iph->pt_destroy_h = pt_destroy;
	iph->init_command_set_h = init_command_set;
	iph->add_command_h = add_command;
	iph->del_command_h = del_command;
	iph->parse_command_h = parse_command;
	iph->parse_configuration_h = parse_configuration;
	iph->interpret_command_h = interpret_command;
	iph->free_cmd_line_h =free_cmd_line;
	iph->h_create_h = h_create;
	iph->h_insert_h = h_insert;
	iph->h_search_h = h_search;
	iph->h_remove_h = h_remove;
	iph->h_destroy_h = h_destroy;
	iph->h_count_h = h_count;
	iph->h_iter_begin_h = h_iter_begin;
	iph->h_iter_end_h = h_iter_end;
	iph->h_iter_next_h = h_iter_next;
	iph->h_iter_getkey_h = h_iter_getkey;
	iph->h_iter_getvalue_h = h_iter_getvalue;
	iph->compare_ipv6_h = compare_ipv6;
	iph->hash_ipv6_h = hash_ipv6;
	iph->set_extent_h = set_extent;
	iph->get_extent_h = get_extent;

	memcpy(iph->handlers, ih, sizeof(handler_t) * HDL_MAX_HANDLERS);
	memcpy(iph->callbacks, icb, sizeof(callback_t*) * CB_MAX_CALLBACKS);

	mnode_plugins = pt_create(PTABLE_TRIMED, 0, NULL);
	if (mnode_plugins == NULL)
	{
		ERR("setup_plugin_manager() - mnode_plugins\n");
		exit(EXIT_FAILURE);
	}

	connectors = pt_create(PTABLE_TRIMED, 0, NULL);
	if (connectors == NULL)
	{
		ERR("setup_plugin_manager() - connectors\n");
		exit(EXIT_FAILURE);
	}

	ciphers = pt_create(PTABLE_TRIMED, 0, NULL);
	if (ciphers == NULL)
	{
		ERR("setup_plugin_manager() - ciphers\n");
		exit(EXIT_FAILURE);
	}

	authenticators = pt_create(PTABLE_TRIMED, 0, NULL);
	if (authenticators == NULL)
	{
		ERR("setup_plugin_manager() - authenticators\n");
		exit(EXIT_FAILURE);
	}

	iph->connectors = connectors;
	iph->ciphers = ciphers;
	iph->authenticators = authenticators;
	iph->mdevice_g = mdevice_g;
}

void deinit_plugin_manager(void)
{
	int idx;

	for (idx = mnode_plugins->length-1; idx >= 0; idx--)
	{
		if (mnode_plugins->ptable[idx] != NULL)
			unload_plugin(((plugin_t *) mnode_plugins->ptable[idx])->name);
	}

	if (iph != NULL)
		free(iph);

	pt_destroy(connectors);
	pt_destroy(ciphers);
	pt_destroy(authenticators);
	pt_destroy(mnode_plugins);
}

int get_connector_id(char *type)
{
	int idx;

	if (strlen(type) == 0)
		return -1;

	for (idx = 0; idx < connectors->length; idx++)
		if (connectors->ptable[idx] != NULL)
			if (strncmp(type, ((connector_t *) connectors->ptable[idx])->type, strlen(type)) == 0)
				break;

	if (idx >= connectors->length)
		return -1;

	return idx;
}

int get_cipher_id(char *type)
{
	int idx;

	if (strlen(type) == 0)
		return -1;

	for (idx = 0; idx < ciphers->length; idx++)
		if (ciphers->ptable[idx] != NULL)
			if (strncmp(type, ((cipher_t *) ciphers->ptable[idx])->type, strlen(type)) == 0)
				break;

	if (idx >= ciphers->length)
		return -1;

	return idx;
}

int get_authenticator_id(char *type)
{
	int idx;

	if (strlen(type) == 0)
		return -1;

	for (idx = 0; idx < authenticators->length; idx++)
		if (authenticators->ptable[idx] != NULL)
			if (strncmp(type, ((authenticator_t *) authenticators->ptable[idx])->type, strlen(type)) == 0)
				break;

	if (idx >= authenticators->length)
		return -1;

	return idx;
}
