/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <ptable.h>


ptable_t *pt_create(ptable_type_t type, int size, valuecmp_f value_cmp_function)
{
	ptable_t *pt_ret = NULL;

	if (((size == 0) && (type & PTABLE_FIXED)) || (size < 0))
		return NULL;

	if (type == (PTABLE_SEARCH | PTABLE_TRIMED))
		return NULL;

	pt_ret = (ptable_t *) calloc(1, sizeof(ptable_t));
	if (pt_ret == NULL)
		return NULL;

	if (size != 0)
	{
		pt_ret->ptable = (void **) calloc(size, sizeof(void *));
		if (pt_ret->ptable == NULL)
		{
			free(pt_ret);
			return NULL;
		}
	}

	pt_ret->length = size;
	pt_ret->type = type;
	pt_ret->value_compare = value_cmp_function;

	return pt_ret;
}

int pt_insert(ptable_t *pt, void *value)
{
	void **n_ptable = NULL;
	int idx;

	for(idx = 0; (idx < pt->length) && (pt->ptable[idx] != NULL); idx++);

	if ((idx == pt->length) && (pt->type & PTABLE_FIXED))
		return -1;

	if (idx == pt->length)
	{
		n_ptable = (void **) realloc(pt->ptable, (pt->length+1) * sizeof(void *));
		if (n_ptable == NULL)
			return -1;

		idx = pt->length;
		pt->length++;
		pt->ptable = n_ptable;
	}

	pt->ptable[idx] = value;
	pt->count++;

	if (pt->type & PTABLE_SEARCH)
		qsort((void *) pt->ptable, pt->count, sizeof(void *), pt->value_compare);

	return idx;
}

void *pt_search(ptable_t *pt, void *value)
{
	void *val_search = value;
	void **ret_val;

	if (!(pt->type & PTABLE_SEARCH))
		return NULL;

	ret_val = (void **) bsearch(&val_search, (const void *) pt->ptable, pt->count, sizeof(void *), pt->value_compare);
	
	if (ret_val == NULL)
		return NULL;

	return *ret_val;
}

void *pt_remove(ptable_t *pt, int index)
{
	void **n_ptable = NULL;
	int n_length = -1;
	void *ret_val = NULL;

	if (index >= pt->length)
		return NULL;

	ret_val = pt->ptable[index];
	pt->ptable[index] = NULL;

	if (pt->type & PTABLE_FIXED)
		n_length = pt->length;
	else
	{
		n_length = pt->length - 1;

		if ((n_length == 0) && !(pt->type & PTABLE_FIXED))
		{
			free(pt->ptable);
			pt->ptable = NULL;
			pt->length = 0;
			pt->count = 0;
			return ret_val;
		}
	}

	if ((pt->type & PTABLE_SEARCH) || ((index == n_length) && (pt->type & PTABLE_TRIMED)))
	{
		if (!(pt->type & PTABLE_FIXED))
			n_ptable = (void **) calloc(n_length, sizeof(ptable_t));

		if (n_ptable != NULL)
			pt->length = n_length;
		else
			n_ptable = pt->ptable;

		if (index > 0)
			memmove(n_ptable, pt->ptable, sizeof(void *) * index);

		if ((index < n_length) && !(pt->type & PTABLE_TRIMED))
			memmove(n_ptable+index, pt->ptable+index+1, sizeof(void *) * (n_length - index));

		if (n_ptable == pt->ptable)
			pt->ptable[pt->length-1] = NULL;
		else
		{
			free(pt->ptable);
			pt->ptable = n_ptable;
		}
	}

	pt->count--;

	if (pt->type & PTABLE_SEARCH)
		qsort((void *) pt->ptable, pt->count, sizeof(void *), pt->value_compare);

	return ret_val;
}

int pt_destroy(ptable_t *pt)
{
	if (pt == NULL)
		return 0;

	if (pt->ptable != NULL)
		free(pt->ptable);

	free(pt);

	return 0;
}
