/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <utils.h>

extern ptable_t *mnode_plugins;
extern mdevice_t *mdevice_g;

mdevice_t *init_mdevice(char *config_file)
{
	mdevice_t *mdevice_ret = NULL;

	mdevice_ret = (mdevice_t *) calloc(1, sizeof(mdevice_t));
	if (mdevice_ret == NULL)
	{
		ERR("init_mdevice()");
		return NULL;
	}

	mdevice_ret->w_buffer = (uint8_t *) calloc(VPZONE_DEFAULT_BUFFER_SIZE, sizeof(uint8_t));
	if (mdevice_ret->w_buffer == NULL)
	{
		free(mdevice_ret);
		ERR("init_mdevice()");
		return NULL;
	}

	mdevice_ret->dns_reqs = pt_create(PTABLE_FIXED | PTABLE_SEARCH, 
									  VPZONE_DEFAULT_MAX_DNS_REQ, dnsreq_cmp);
	if (mdevice_ret->dns_reqs == NULL)
	{
		free(mdevice_ret->w_buffer);
		free(mdevice_ret);
		ERR("init_mdevice()");
		return NULL;
	}

	mdevice_ret->timeout.tv_sec = VPZONE_DEFAULT_TIMEOUT;
	mdevice_ret->prio_ipv4 = true;
	mdevice_ret->buffer_size = VPZONE_DEFAULT_BUFFER_SIZE;

	strncpy(mdevice_ret->config_file, config_file, strlen(config_file));

	mdevice_ret->mnodes_limit = DEFAULT_NODE_LIMIT;

	mdevice_ret->authenticator_id = -1;

	return mdevice_ret;
}

mnode_t *init_mnode(void)
{
	mnode_t *mnode_ret = NULL;

	mnode_ret = (mnode_t *) calloc(1, sizeof(mnode_t));
	if (mnode_ret == NULL)
	{
		ERR("init_mnode()");
		return NULL;
	}

	mnode_ret->r_buffer = (uint8_t *) calloc(mdevice_g->buffer_size, sizeof(uint8_t));
	if (mnode_ret->r_buffer == NULL)
	{
		ERR("init_mnode()");
		free(mnode_ret);
		return NULL;
	}

	mnode_ret->w_buffer = (uint8_t *) calloc(mdevice_g->buffer_size, sizeof(uint8_t));
	if (mnode_ret->w_buffer == NULL)
	{
		ERR("init_mnode()");
		free(mnode_ret->r_buffer);
		free(mnode_ret);
		return NULL;
	}

	mnode_ret->extents = pt_create(PTABLE_TRIMED, mnode_plugins->length, NULL);
	if (mnode_ret->extents == NULL)
	{
		free(mnode_ret->r_buffer);
		free(mnode_ret->w_buffer);
		free(mnode_ret);
		return NULL;
	}

	mnode_ret->is_server = false;

	init_mnode_extent(mnode_ret);

	mnode_ret->connector_id = -1;
	mnode_ret->cipher_id = -1;
	mnode_ret->authenticator_id = mdevice_g->authenticator_id;

	return mnode_ret;
}

mserver_t *init_mserver()
{
	mserver_t *mserver_ret = NULL;

	mserver_ret = (mserver_t *) calloc(1, sizeof(mserver_t));
	if (mserver_ret == NULL)
	{
		ERR("init_mserver()");
		return NULL;
	}

	mserver_ret->extents = pt_create(PTABLE_TRIMED, mnode_plugins->length, NULL);
	if (mserver_ret->extents == NULL)
	{
		ERR("init_mserver()");
		free(mserver_ret);
		return NULL;
	}

	init_mserver_extent(mserver_ret);

	mserver_ret->connector_id = -1;
	mserver_ret->cipher_id = -1;
	mserver_ret->authenticator_id = mdevice_g->authenticator_id;

	return mserver_ret;
}

int init_mdevice_hash(mdevice_t *mdevice)
{
	mdevice->ip_mnodes = h_create(mdevice->mnodes_limit, hash_ipv6, compare_ipv6);
	if (mdevice->ip_mnodes == NULL)
		return -1;

	return 0;
}

void free_extents(ptable_t *extents)
{
	int i;

	if (extents == NULL)
		return;

	for (i = 0; i < extents->length; i++)
		if (extents->ptable[i] != NULL)
			free(extents->ptable[i]);
}

void free_mnode(mnode_t *mnode)
{
	free(mnode->r_buffer);
	free(mnode->w_buffer);

	remove_mnode_extent(mnode);

	free_extents(mnode->extents);
	pt_destroy(mnode->extents);

	free(mnode);
}

void free_mserver(mserver_t *mserver)
{
	remove_mserver_extent(mserver);

	free_extents(mserver->extents);
	pt_destroy(mserver->extents);

	free(mserver);
}

void free_mdevice(mdevice_t *mdevice)
{
	pt_destroy(mdevice->dns_reqs);
	free(mdevice->w_buffer);
	free(mdevice);
}

int set_extent(bool m_type, void *mthing, int p_id, extent_type_t p_type, extent_t value)
{
	ptable_t *extents = NULL;
	int i;

	if (p_type >= MAX_EXTENT_TYPE)
		return -1;

	if (m_type)
		extents = ((mnode_t *)mthing)->extents;
	else
		extents = ((mserver_t *)mthing)->extents;
	
	if (extents->length < mnode_plugins->length)
	{
		for (i = extents->length; i < mnode_plugins->length; i++)
			if (pt_insert(extents, NULL) < 0)
				return -1;
	}

	if (extents->ptable[p_id] == NULL)
	{
		extents->ptable[p_id] = (extent_t *) calloc(MAX_EXTENT_TYPE, sizeof(extent_t));
		if (extents->ptable[p_id] == NULL)
			return -1;
	}

	((extent_t *) extents->ptable[p_id])[p_type] = value;

	return 0;
}

extent_t get_extent(bool m_type, void *mthing, int p_id, extent_type_t p_type)
{
	ptable_t *extents = NULL;

	if (p_type >= MAX_EXTENT_TYPE)
		return NULL;

	if (m_type)
		extents = ((mnode_t *)mthing)->extents;
	else
		extents = ((mserver_t *)mthing)->extents;
	
	if (extents->length < p_id)
		return NULL;

	if (extents->ptable[p_id] == NULL)
		return NULL;

	return ((extent_t *) extents->ptable[p_id])[p_type];
}
int parse_connector_line(char *src_line, int *conn_id, int *ciph_id, char *host, uint16_t *port)
{
	char *p_sep;
	char *sep;
	uint32_t port_test;
	bool host_notfound = true;

	if (strlen(src_line) < 5)
		return -1;

	*port = VPZONE_DEFAULT_PORT;
	p_sep = sep = src_line + strlen(src_line);

	while (host_notfound && (sep > src_line))
	{
		if ((sep = strrchr(src_line, ':')) == NULL)
			return -2;

		if ((strchr(sep, '.') == NULL) && (strchr(sep, ']') == NULL))
		{
			errno = 0;
			port_test = (uint32_t) strtol(sep+1, (char **)NULL, 10);
			if ((errno != ERANGE) && (port_test < 65535) && (port_test > 0))
			{
				*port = (uint16_t) port_test;
				*sep = '\0';
				p_sep = sep;
				continue;
			}
		}

		if (strchr(sep, ']') != NULL)
		{
			if ((sep = strrchr(src_line, '[')) == NULL)
				return -3;
			p_sep -= 2;
		}

		strncpy(host, sep+1, p_sep - sep);
		host_notfound = false;
		if (*sep == '[')
			sep--;
	}

	*sep = '\0';
	p_sep = sep;

	if ((sep = strrchr(src_line, ':')) == NULL)
		return -4;

	if ((*ciph_id = get_cipher_id(sep + 1)) < 0)
		return -5;

	*sep = '\0';

	if ((*conn_id = get_connector_id(src_line)) < 0)
		return -6;

	return 0;
}

int parse_address_line(char *src_line, char *address, int *prefix_len)
{
	char *px;

	if ((px = strchr(src_line, '/')) == NULL)
		return -1;

	*prefix_len = atol(px+1);
	if ((*prefix_len < 0) || (*prefix_len > 128))
		return -2;
	*px = '\0';

	if ((px - src_line) > INET6_ADDRSTRLEN)
		return -3;

	strncpy(address, src_line, INET6_ADDRSTRLEN);

	return 0;
}

int compare_ipv6(void *c1, void *c2)
{
	int i;

	struct in6_addr *ip1 = (struct in6_addr *) c1;
	struct in6_addr *ip2 = (struct in6_addr *) c2;

	for (i = 0; i < 4; i++)
	{
		if (((const uint32_t *) (ip1))[i] < ((const uint32_t *) (ip2))[i])
			return -1;

		if (((const uint32_t *) (ip1))[i] > ((const uint32_t *) (ip2))[i])
			return 1;
	}

	return 0;
}

uint32_t hash_ipv6(void *key)
{
	struct in6_addr *ip = (struct in6_addr *) key;
	int i;
	uint32_t hash = 5381;

	for (i = 0; i < 4; i++)
		hash = ((hash << 5) + hash) + ((const uint32_t *) (ip))[i];

	return hash;
}

int is_ipv6_unicast(void *c1, void *c2)
{
	c2 = NULL;

	if (!IN6_IS_ADDR_LINKLOCAL((struct in6_addr *) c1))
		return -1;

	return 0;
}
