/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <udp.h>

PLUGIN_DEPENDANCIES(NULL);

PLUGIN_GBL_VARS;

connector_t udp[] = {
	{
		-1,
		"udp",
		connect_mnode_udp,
		disconnect_mnode_udp,
		setup_mserver_udp,
		accept_mnode_udp,
		close_mserver_udp
	},
	{-1, "", NULL, NULL, NULL, NULL, NULL }
};

plugin_type_t init(int plugin_id, in_plugin_handlers_t *iph)
{
	PLUGIN_SET_VARS(plugin_id, iph);

	udp[0].plugin_id = plugin_id;

	log_msg(INFO_LVL, "Loaded UDP plugin\n");

	return PLUGIN_CONNECTOR;
}

int unload(void)
{
	return 0;
}

connector_t *get_connectors(void)
{
	return udp;
}

int setup_mserver_udp(mdevice_t *mdevice, mserver_t *mserver)
{
	int flags;

	if (mserver == NULL)
		return 0;

	if ((mserver->sock = socket(mserver->address.sa_family, SOCK_DGRAM, 0)) < 0)
	{
		ERR("setup_mserver_udp()");
		return -1;
	}

	/*Setting non blocking mode*/
	if ((flags = fcntl(mserver->sock, F_GETFL, 0)) < 0)
        flags = 0;

    if (fcntl(mserver->sock, F_SETFL, flags | O_NONBLOCK) < 0)
	{
		ERR("setup_mserver_udp()");
		close(mserver->sock);
		return -1;
	}

	/*Reuse address*/
	flags = 1;
	if(setsockopt(mserver->sock, SOL_SOCKET, SO_REUSEADDR, &flags, sizeof(int)) < 0)
		ERR("setup_mserver_udp()");

	/*Binding address and port*/
	((struct sockaddr_in *) &(mserver->address))->sin_port = htons(mserver->port);

	if (bind(mserver->sock, &(mserver->address), sizeof(struct sockaddr)) < 0) 
	{
		ERR("setup_mserver_udp()");
		close(mserver->sock);
		return -1;
	}

	DEBUG_MSG("UDP Server running on %s:%d\n", mserver->host, mserver->port);

	if (mdevice->mservers != NULL)
		mserver->next = mdevice->mservers;

	mdevice->mservers = mserver;

	/*Adding it in event base*/
	DEBUG_MSG("Adding UDP server event\n");
	event_set(&mserver->server_event, mserver->sock, EV_READ|EV_PERSIST, accept_mnode_udp, mdevice);

	event_add(&mserver->server_event, NULL);

	return 0;
}

int close_mserver_udp(mdevice_t *mdevice, mserver_t *mserver)
{
	if (mdevice == NULL || mserver == NULL)
		return -1;

	return 0;
}

void accept_mnode_udp(int sock, short event, void *arg)
{
	char trans_buff[512];
	struct sockaddr mnode_addr;
	mnode_t *mnode = NULL;
	socklen_t addrlen;
	int flags;
	int ret_code;
	mdevice_t *mdevice = (mdevice_t*) arg;
	mserver_t *mserver = NULL;

	/*Retrieve mserver*/
	for (mserver = mdevice->mservers; 
		 mserver != NULL && mserver->sock != sock; 
		 mserver = mserver->next);

	/*Make compiler happy*/
	if (event == EV_READ)
		log_msg(WARNING_LVL, "accept_udp_mnode event is not the one we want\n");
	/***/

	addrlen = sizeof(mnode_addr);

	if (recvfrom(mserver->sock, trans_buff, 512, 0, &mnode_addr, &addrlen) < 0)
		return;

	if (strncmp((const char*)trans_buff, "MNODE_UDP_CONNECT", 18) != 0)
		return;

	if((mnode = init_mnode()) == NULL)
	{
		log_msg(ERROR_LVL, "Can't accept new connections due to lack of memory!\n");
		return;
	}

	mnode->is_server = true;

	if ((mnode->sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		ERR("accept_mnode_udp()");
		free_mnode(mnode);
		return;
	}

	if (connect(mnode->sock, &mnode_addr, sizeof(struct sockaddr_in)) < 0)
	{
		ERR("accept_mnode_udp()");
		close(mnode->sock);
		free_mnode(mnode);
		return;
	}

	/*Optimize that (make it event based and manage timeouts)*/
	send(mnode->sock, "MNODE_UDP_OK", 13, 0);

	recv(mnode->sock, trans_buff, 512, 0);

	if (strncmp((const char*)trans_buff, "MNODE_UDP_OK", 13) != 0)
	{
		close(mnode->sock);
		free_mnode(mnode);
		return;
	}

	/*To make async*/
	if(getnameinfo(&mnode_addr, sizeof(mnode_addr), mnode->host, 512, NULL, 0, NI_NUMERICHOST) < 0)
	{
		ERR("accept_mnode_udp()");
		close(mnode->sock);
		free_mnode(mnode);
		return;
	}

	mnode->port = htons(((struct sockaddr_in *) &mnode_addr)->sin_port);

	mnode->connector_id = mserver->connector_id;
	mnode->cipher_id = mserver->cipher_id;
	mnode->authenticator_id = mserver->authenticator_id;

	setup_mnode_server_cipher(ret_code, mnode);
	if (ret_code < 0)
	{
		close(mnode->sock);
		free_mnode(mnode);
		return;
	}

	setup_mnode_server_authentification(ret_code, mnode);
	if (ret_code < 0)
	{
		close(mnode->sock);
		free_mnode(mnode);
		return;
	}

	/*Setting non blocking mode*/
	if ((flags = fcntl(mnode->sock, F_GETFL, 0)) < 0)
        flags = 0;

    if (fcntl(mnode->sock, F_SETFL, flags | O_NONBLOCK) < 0)
	{
		ERR("accept_mnode_udp()");
		close(mnode->sock);
		free_mnode(mnode);
		return;
		}

	log_msg(INFO_LVL, "A node has connected %s:%d\n", mnode->host, mnode->port);

	post_accept(ph->callbacks[CB_POST_ACCEPT], mdevice, mnode);

	/*Set events and add read one*/
	event_set(&mnode->read_event, mnode->sock, EV_READ|EV_PERSIST, recv_mnode_h(mnode), mnode);
	event_set(&mnode->write_event, mnode->sock, EV_WRITE|EV_PERSIST, send_mnode_h(mnode), mnode);
	event_add(&mnode->read_event, NULL);

	post_accept(ph->callbacks[CB_POST_ACCEPT], mdevice, mnode);
}

int connect_mnode_udp(mnode_t *mnode)
{
	char trans_buff[512];
	int flags;
	socklen_t addrlen;
	int ret_code;

	if ((mnode->sock = socket(mnode->address.sa_family, SOCK_DGRAM, 0)) < 0)
	{
		ERR("connect_mnode_udp()");
		return -1;
	}

	((struct sockaddr_in *) &(mnode->address))->sin_port = htons(mnode->port);

	sendto(mnode->sock, "MNODE_UDP_CONNECT", 18, 0, &(mnode->address), sizeof(struct sockaddr));
	
	addrlen = sizeof(struct sockaddr);
	recvfrom(mnode->sock, trans_buff, 512, 0, &(mnode->address), &addrlen);

	if (strncmp((const char*)trans_buff, "MNODE_UDP_OK", 13) != 0)
	{
		close(mnode->sock);
		return -1;
	}

	if (connect(mnode->sock, &(mnode->address), sizeof(struct sockaddr_in)) < 0)
	{
		ERR("connect_mnode_udp()");
		close(mnode->sock);
		return -1;
	}

	send(mnode->sock, "MNODE_UDP_OK", 13, 0);

	setup_mnode_client_cipher(ret_code, mnode);
	if (ret_code < 0)
	{
		DEBUG_MSG("setup_mnode_client_cipher failed\n");
		close(mnode->sock);
		return -1;
	}

	mnode->authenticator_id = mdevice_g->authenticator_id;

	setup_mnode_client_authentification(ret_code, mnode);
	if (ret_code < 0)
	{
		DEBUG_MSG("setup_mnode_client_authentification failed\n");
		close(mnode->sock);
		return -1;
	}

	/*Setting non blocking mode*/
	if ((flags = fcntl(mnode->sock, F_GETFL, 0)) < 0)
        flags = 0;

    if (fcntl(mnode->sock, F_SETFL, flags | O_NONBLOCK) < 0)
	{
		ERR("connect_mnode_udp()");
		close(mnode->sock);
		return -1;
	}

	log_msg(INFO_LVL, "Connected to mnode %s:%d\n", mnode->host, mnode->port);

	post_connect(ph->callbacks[CB_POST_CONNECT], mdevice_g, mnode);

	/*Set events and add read one*/
	event_set(&mnode->read_event, mnode->sock, EV_READ|EV_PERSIST, recv_mnode_h(mnode), mnode);
	event_set(&mnode->write_event, mnode->sock, EV_WRITE|EV_PERSIST, send_mnode_h(mnode), mnode);
	event_add(&mnode->read_event, NULL);

	return 0;
}

void disconnect_mnode_udp(mnode_t *mnode)
{
	if (mnode == NULL)
		return;
}
