/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <babeld.h>

static int setup_babeld(void);
static int setup_netlink_socket(uint32_t groups);
static int netlink_recv(int sock, char *buffer, int buff_len);
static void netlink_watch(int sock, short event, void *null);

PLUGIN_DEPENDANCIES(NULL);

PLUGIN_GBL_VARS;

char *babeld_params = NULL;
int hashsize = 20;

static pid_t babel_proc = -1;

static int nlwatch_socket = -1;

static void *hm_gateways = NULL;

static struct event netlink_event;

static const command_t babel_commands[] = {
	{ "babeld-command", CMD_NONE, babeld_command_cmd, "" },
	{ "babeld-hashsize", CMD_NONE, babeld_hashsize_cmd, "" },
	{ "", CMD_NONE, NULL, "" }
};

plugin_type_t init(int plugin_id, in_plugin_handlers_t *iph)
{
	ptable_t *p_bcmd = NULL;

	PLUGIN_SET_VARS(plugin_id, iph);

	p_bcmd = init_command_set(babel_commands);
	if (p_bcmd == NULL)
		return -1;

	parse_configuration(mdevice_g->config_file, p_bcmd);

	pt_destroy(p_bcmd);

	if ((nlwatch_socket = setup_netlink_socket(RTMGRP_IPV6_ROUTE)) < 0)
		return -1;

	/*Set hashmap*/
	if ((hm_gateways = h_create(hashsize, hash_ipv6, compare_ipv6)) == NULL)
		return -1;

	/*Set netlink_watch in event pool*/
	DEBUG_MSG("Adding Babeld netlink watchdog event\n");
	event_set(&netlink_event, nlwatch_socket, EV_READ|EV_PERSIST, netlink_watch, NULL);

	event_add(&netlink_event, NULL);

	if ((setup_babeld() < 0) ||
		(set_handler(HDL_SELECT_ROUTE, select_route_babeld) != HOOK_SUCCESSFUL))
		return -1;

	log_msg(INFO_LVL, "Loaded BABELD router plugin\n");

	return PLUGIN_DEFAULT;
}

int unload(void)
{
	if (babeld_params != NULL)
		free(babeld_params);

	if (hm_gateways != NULL)
	{
		h_iter_begin(hm_gateways);

		while(h_iter_next(hm_gateways))
			free(h_remove(hm_gateways, h_iter_getkey(hm_gateways), true));

		h_iter_end(hm_gateways);

		h_destroy(hm_gateways);
	}
	
	event_del(&netlink_event);

	if (nlwatch_socket != -1)
		close(nlwatch_socket);

	if (babel_proc > 0)
		kill(babel_proc, SIGTERM);

	unset_handler(HDL_SELECT_ROUTE);

	return 0;
}

static int setup_babeld(void)
{
	char **cmd;

	if ((cmd = parse_command(babeld_params)) == NULL)
	{
		log_msg(ERROR_LVL, "Could not parse babeld command line.");
		return -1;
	}

	babel_proc = fork();
	switch (babel_proc)
	{
	case -1:
		ERR("setup_babeld()");
		return -1;
		break;
	case 0:
		execv(*cmd, cmd);
		ERR("setup_babeld()");
		exit(EXIT_FAILURE);
	default:
		break;
	}

	free(babeld_params);
	babeld_params = NULL;

	free_cmd_line(cmd);

	return 0;
}

static int setup_netlink_socket(uint32_t groups)
{
	struct sockaddr_nl nl_addr;
	int nl_sock;

	if ((nl_sock = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE)) < 0)
	{
		ERR("setup_netlink_socket()");
		return -1;
	}

	if (groups != 0)
	{
		memset((void *) &nl_addr, 0, sizeof(nl_addr));
		
		nl_addr.nl_family = AF_NETLINK;
		nl_addr.nl_pid = getpid();
		nl_addr.nl_groups = groups;
		
		if (bind(nl_sock, (struct sockaddr *)&nl_addr, sizeof(nl_addr)) < 0)
		{
			ERR("setup_netlink_socket()");
			close(nl_sock);
			return -1;
		}
	}

	return nl_sock;
}

static int netlink_recv(int sock, char *buffer, int buff_len)
{
	struct sockaddr_nl nladdr;
	struct msghdr msg;
	struct iovec iov;

	memset(buffer, 0, buff_len);
	
	iov.iov_base = (void *)buffer;
	iov.iov_len = buff_len;
	
	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;
	
	memset(&msg, 0, sizeof(msg));
	msg.msg_name = (void *)&(nladdr);
	msg.msg_namelen = sizeof(nladdr);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	
	return (int) recvmsg(sock, &msg, 0);
}

static void netlink_watch(int sock, short event, void *null)
{
	char buffer[8192];
	ssize_t ret;
	int len;
	struct nlmsghdr *nlh;
	struct rtmsg *rtm;
	struct rtattr *rta;
	bool addroute = false;
	struct in6_addr *ip6_distant = NULL;
	struct in6_addr *ip6_gateway = NULL;
	struct in6_addr *ip6_oldgw = NULL;

	/*make compiler happy*/
	if ((event != EV_READ) || (null != NULL))
		log_msg(WARNING_LVL, "Babeld - netlink_watch(): bad event?\n");
    /**/

	if ((ret = netlink_recv(sock, buffer, sizeof(buffer))) <= 0)
		return;

	nlh = (struct nlmsghdr *)buffer;
	if((!NLMSG_OK(nlh, (unsigned int) ret)) ||
	   (nlh->nlmsg_type == NLMSG_ERROR))
	{
		DEBUG_MSG("Babeld - netlink_watch(): NOT NLMSG_OK or NLMSG_ERROR\n");
		return;
	}

	if (nlh->nlmsg_type == RTM_NEWROUTE)
		addroute = true;
	else
		if (nlh->nlmsg_type != RTM_DELROUTE)
			return;
	
	rtm = (struct rtmsg *) NLMSG_DATA(nlh);
	rta = (struct rtattr *) RTM_RTA(rtm);
		
	len = NLMSG_PAYLOAD(nlh, ret) + ret;

	while(RTA_OK(rta, len))
	{
		if (rta->rta_type == RTA_DST)
		{
			DEBUG_MSG("Babeld - netlink_watch(): got RTA_DST\n");
			ip6_distant = (struct in6_addr *) calloc(1, sizeof(struct in6_addr));
			if (ip6_distant == NULL)
			{
				ERR("netlink_watch()");
				if (ip6_gateway != NULL)
					free (ip6_gateway);
				return;
			}
			memcpy(ip6_distant, RTA_DATA(rta), sizeof(struct in6_addr));
		}

		if(rta->rta_type == RTA_GATEWAY)
		{
			DEBUG_MSG("Babeld - netlink_watch(): got RTA_GATEWAY\n");
			ip6_gateway = (struct in6_addr *) calloc(1, sizeof(struct in6_addr));
			if (ip6_gateway == NULL)
			{
				ERR("netlink_watch()");
				if (ip6_distant != NULL)
					free (ip6_distant);
				return;
			}
			memcpy(ip6_gateway, RTA_DATA(rta), sizeof(struct in6_addr));
			break;
		}
		
		rta = RTA_NEXT(rta, len);
	}

	/*update route*/
	/**************/

	/*is distant ip a UNICAST one? if so: we don't care*/
	/*Check if not neighbour*/
	/*Check if gateway is an existing neighbour*/
	if ((ip6_distant == NULL) || (ip6_gateway == NULL) ||
	   IN6_IS_ADDR_LINKLOCAL(ip6_distant) ||
	   (h_search(mdevice_g->ip_mnodes, ip6_distant) != NULL) ||
	   (h_search(mdevice_g->ip_mnodes, ip6_gateway) == NULL))
	{
		DEBUG_MSG("Babeld - netlink_watch(): useless notification\n");
		if (ip6_distant != NULL)
			free(ip6_distant);
		if (ip6_gateway != NULL)
			free(ip6_gateway);
		return;
	}

	/*If a route for this distant ip exists, let's remove it*/
	ip6_oldgw = h_remove(hm_gateways, ip6_distant, true);
	
	if (ip6_oldgw != NULL)
		free(ip6_oldgw);
	
	/*We inject the route (if only RTM_NEWROUTE)*/
	if (addroute)
		h_insert(hm_gateways, ip6_distant, ip6_gateway);
	else
		free(ip6_gateway);
}

mnode_t *select_route_babeld(struct in6_addr *ip6_addr)
{
	struct in6_addr *ip6_gateway = NULL;
	mnode_t *mnode_gw = NULL;

	/*Is ip6_addr in hm_gateways?*/
	if ((ip6_gateway = (struct in6_addr *) h_search(hm_gateways, (void *) ip6_addr)) == NULL)
	{
		DEBUG_MSG("Babeld - select_route_babeld(): Route not found, should ask kernel later\n");
		return NULL;
	}

	mnode_gw = (mnode_t *) h_search(mdevice_g->ip_mnodes, (void *) ip6_gateway);

	return mnode_gw;
}
